unit SharedMemory;

{
  Unit SharedMemory:

  Provides structured access to shared memory between processes. We have
  six basic methods for accessing shared memory, which use Mutex objects
  for synchronization. They allow us to write a value at a specific offset
  from the starting address of the view of the mapped file, read the value
  at a offset, flush the contents to the shared file (if any) and write a
  value at a offset, then flush the contents to the shared file. They also
  allow writing an entire block with the same value, and writing and flushing

  Usage is explained along in the code.
}

  Interface



USES
  Windows, Messages, SysUtils;

CONST
  DefaultMemSize = 512;

TYPE
  ESharedMemoryError = Class(Exception);
  
  TArrayOfBytes = ARRAY[1..MAXINT] OF Byte;
  PArrayOfBytes = ^TArrayOfBytes;

  TSharedMemory = Class
    Private
      FHndFile,             { Handle to the file object }
      FHndFileMapping,      { Handle to the file-mapping object }
      FHndMutex: THandle;   { Handle to the mutex object }

      FStartingDirFileView: Pointer;  { Points to the starting address of the file-view }

      FName,                { Name of the memory }
      FMutexName: String;   { Name of the mutex object }
      FLastError: Integer;  { Obtained by GetLastError }
      FFirst,               { Am I the creator of the shared memory? }
      FCreated: Boolean;    { Is shared memory created? }
      FMemSize: DWord;      { Size of the memory block: by default, 512 bytes }

      FP: PArrayOfBytes;    { Pointer to the array of available shared bytes of memory }

      FUNCTION SM_GetLastError: Integer;
      FUNCTION GetMemName: String;
      FUNCTION GetMemSize: DWord;

    Public
      Constructor Create(smName,              { Name of the memory }
                         smMutexName: String; { Name of the mutex object }
                         smSize: DWord;       { Size of the memory }
                         smPSecurityAttributes: PSecurityAttributes;  { Security attr. }
                         smFile: String);     { Name of physical file for sharing memory,
                                                if any, or empty string if none }
      Destructor  Destroy; Override;

      { These properties give us information about the state of the
        shared memory }

      Property LastError: Integer Read SM_GetLastError;
      Property MemName:   String  Read GetMemName;
      Property MemSize:   DWord   Read GetMemSize;
      Property IsCreated: Boolean Read FCreated;
      Property AmIFirst:  Boolean Read FFirst;

      { And these methods allow us access shared memory }

      { Individual values... }

      PROCEDURE WriteValueFlush(Offset: Integer; Value: Byte);
      PROCEDURE WriteValue(Offset: Integer; Value: Byte);
      FUNCTION  ReadValue(Offset: Integer): Byte;

      { Blocks... }

      PROCEDURE WriteBlock(Offset: Integer; Count: DWord; Value: Byte);
      PROCEDURE WriteBlockFlush(Offset: Integer; Count: DWord; Value: Byte);

      { And write all of them to file (if exists) }

      PROCEDURE Flush;
  END;




  Implementation



  
{
  Ex:

  VAR
    SM: TSharedMemory;

  [...]

  SM := TSharedMemory.Create('My shared memory name',
                             'My mutex name',
                             1024,            // Size of shared memory
                             Nil,             // Default security attributes
                             '');             // Use the page file
  [... use it ...]

  SM.WriteValue(13, 65);         // Write value 65 at address (FStartingDirFileWiew + 12)
  Variable := SM.ReadValue(17);  // Read value at address (FStartingDirFileWiew + 16)
  SM.Flush;                      // Flush contents of shared memory into the file
                                 // specified by the parameter smFile, if smFile <> '',
                                 // else it does nothing
  SM.WriteValueFlush(14, 66);    // Write value 66 at address (FStartingDirFileView + 13)
                                 // and then flush the view of the mapped file to the
                                 // physical file (it does not flush anything if
                                 // smFile = '', i.e., when FHndFile = INVALID_HANDLE_VALUE)
  SM.WriteBlock(2, 10, 64);      // Fill addresses (FStartingDirFileView + 1) to
                                 // (FStartingDirFileView + 1 + 10) with value 64
  SM.WriteBlockFlush(2, 10, 64); // The same, but now flushing it to file if any...

  [... continue using it ...]

  SM.Free;
}

{
  You should protect calling to these functions on a TRY ... EXCEPT or
  TRY ... FINALLY block, because any of the following functions raises
  an exception if shared memory is not created:

  PROCEDURE TSharedMemory.WriteValueFlush(Offset: Integer; Value: Byte);
  PROCEDURE TSharedMemory.WriteValue(Offset: Integer; Value: Byte);
  FUNCTION  TSharedMemory.ReadValue(Offset: Integer): Byte;
  PROCEDURE TSharedMemory.Flush;
  PROCEDURE TSharedMemory.WriteBlock(Offset: Integer; Count: DWord; Value: Byte);
  PROCEDURE TSharedMemory.WriteBlockFlush(Offset: Integer; Count: DWord; Value: Byte);

  Shared memory is not created if one of the following fails:

  - Creating the handle for the file object (if a file is specified in constructor
  - Creating the handle for the file-mapping object
  - Creating the view of the file-mapping
  - Creating the handle for the mutex object
}

Constructor TSharedMemory.Create(smName, smMutexName: String;
  smSize: DWord;
  smPSecurityAttributes: PSecurityAttributes;
  smFile: String);

VAR
  OperationOK: Boolean;
BEGIN
  Inherited Create;

  FHndFile :=        INVALID_HANDLE_VALUE;
  FHndFileMapping := INVALID_HANDLE_VALUE;
  FHndMutex :=       INVALID_HANDLE_VALUE;

  FStartingDirFileView := Nil;

  IF smSize = 0 THEN
    FMemSize := DefaultMemSize
  ELSE
    FMemSize := smSize;

  FName := smName;
  FMutexName := smMutexName;

  FCreated := False;
  FFirst := False;

  OperationOK := True;

  { Entering a REPEAT ... UNTIL True; block: we follow steps detailed in MSDN,
    and if any of them fails, then we break execution, getting FCreated = False.
    Calling methods of TSharedMemory with FCreated = False raise an
    exception of class ESharedMemoryError }

  REPEAT
    IF smFile <> '' THEN
    BEGIN
      FHndFile := CreateFile(PChar(smFile),                      { Filename }
                             GENERIC_READ + GENERIC_WRITE,       { Desired access to file }
                             FILE_SHARE_READ + FILE_SHARE_WRITE, { Share mode }
                             smPSecurityAttributes,              { Security attributes }
                             OPEN_ALWAYS,                        { Creation disposition }
                             FILE_ATTRIBUTE_NORMAL,              { Flags and attributes }
                             0);                                 { Handle to template file }

      FLastError := GetLastError;
      OperationOK := (FHndFile <> INVALID_HANDLE_VALUE)
                     AND (
                         (FLastError = ERROR_SUCCESS) OR
                         (FLastError = ERROR_ALREADY_EXISTS)
                         );
    END;

    IF NOT OperationOK AND (FHndFile <> INVALID_HANDLE_VALUE) THEN
      Break; 

    { If we have FHndFile = INVALID_HANDLE_VALUE, then the system will be
      the administrator of the shared memory, by using the page file }

    FHndFileMapping := CreateFileMapping(FHndFile,              { Handle of the file object }
                                         smPSecurityAttributes, { Security attributes }
                                         PAGE_READWRITE,        { Access mode }
                                         0,                     { High size of memory }
                                         FMemSize,              { Low size of memory }
                                         PChar(FName));         { Name of the memory }

    FLastError := GetLastError;
    OperationOK := (FHndFileMapping <> INVALID_HANDLE_VALUE)
                   AND (
                       (FLastError = ERROR_SUCCESS) OR
                       (FLastError = ERROR_ALREADY_EXISTS)
                       );

    IF NOT OperationOK THEN Break;

    { Note that getting ERROR_ALREADY_EXISTS is NOT an error: it only
      tells us that there is someone who has created the object before
      than us (it could be the same process executed twice or more) }

    FFirst := (FLastError <> ERROR_ALREADY_EXISTS) AND (FLastError = ERROR_SUCCESS);

    FStartingDirFileView := MapViewOfFile(FHndFileMapping, { Handle of the file-mapping object }
                                          FILE_MAP_WRITE,  { Access mode }
                                          0,               { High(File Offset in file-mapping) }
                                          0,               { Low(File Offset in file-mapping) }
                                          0);              { All the file is mapped }

    OperationOK := Assigned(FStartingDirFileView);

    IF NOT OperationOK THEN
      Break;

    FHndMutex := CreateMutex(smPSecurityAttributes, { Security attributes for mutex }
                             True,                  { Request inmediate ownership }
                             PChar(FMutexName));    { Name for the mutex (also shared) }

    FLastError := GetLastError;
    OperationOK := (FHndMutex <> INVALID_HANDLE_VALUE)
                   AND (
                       (FLastError = ERROR_SUCCESS) OR
                       (FLastError = ERROR_ALREADY_EXISTS)
                       );

    FCreated := OperationOK;

    { If everything has gone well up to now, then we have shared memory }

    IF FCreated THEN
      FP := FStartingDirFileView;  { and this is the starting address for shared memory }
      
  UNTIL True;
END;

Destructor TSharedMemory.Destroy;
BEGIN
  { Unmap view of file created, then, close all created handles.
    Remember that CloseHandle can raise an exception if an invalid
    handle is passed or if a handle is closed twice }

  TRY
    CloseHandle(FHndFile);
  EXCEPT  { Erase exception }
  END;

  UnmapViewOfFile(FStartingDirFileView);

  TRY
    CloseHandle(FHndFileMapping);
  EXCEPT
  END;

  TRY
    CloseHandle(FHndMutex);
  EXCEPT
  END;

  Inherited Destroy;
END;

FUNCTION TSharedMemory.SM_GetLastError: Integer;
BEGIN
  FLastError := GetLastError;
  Result := FLastError;
END;

{ These functions provide information about the shared memory }

FUNCTION TSharedMemory.GetMemName: String;
BEGIN
  IF FCreated THEN
    Result := FName
  ELSE
  BEGIN
    SetLastError(ERROR_ACCESS_DENIED);
    Result := '';
  END;
END;

FUNCTION TSharedMemory.GetMemSize: DWord;
BEGIN
  IF FCreated THEN
    Result := FMemSize
  ELSE
  BEGIN
    SetLastError(ERROR_ACCESS_DENIED);
    Result := 0;
  END;
END;


{ These are the method that allow read/write access to shared memory }


{ Let's write some value and save it to file... }

PROCEDURE TSharedMemory.WriteValueFlush(Offset: Integer; Value: Byte);
VAR
  dwResultWaitMutex: DWord;
BEGIN
  IF FCreated THEN
  BEGIN
    dwResultWaitMutex := WaitForSingleObject(FHndMutex, INFINITE);

    IF dwResultWaitMutex = WAIT_OBJECT_0 THEN
    BEGIN
      FP^[Offset] := Value;

      IF FHndFile <> INVALID_HANDLE_VALUE THEN
        IF NOT FlushViewOfFile(FStartingDirFileView, Offset) THEN
          RAISE
            ESharedMemoryError.Create('Unexpected error flushing memory to file');

      ReleaseMutex(FHndMutex);
    END;
  END
  ELSE
  BEGIN
    RAISE
      ESharedMemoryError.Create('Can''t write value on a not created shared memory');
    SetLastError(ERROR_ACCESS_DENIED);
  END;
END;

{ Let's write some value... }

PROCEDURE TSharedMemory.WriteValue(Offset: Integer; Value: Byte);
VAR
  dwResultWaitMutex: DWord;
BEGIN
  IF FCreated THEN
  BEGIN
    dwResultWaitMutex := WaitForSingleObject(FHndMutex, INFINITE);

    IF dwResultWaitMutex = WAIT_OBJECT_0 THEN
    BEGIN
      FP^[Offset] := Value;

      ReleaseMutex(FHndMutex);
    END;
  END
  ELSE
  BEGIN
    RAISE
      ESharedMemoryError.Create('Can''t write value on a not created shared memory');
    SetLastError(ERROR_ACCESS_DENIED);
  END;
END;

{ Remember to call GetLastError after calling this function, because if
  Result = 0, this value should be obtained by an error:
  ERROR_ACCESS_DENIED (shared memory is not created)
  or ERROR_OPEN_FAILED (can't get ownership of the mutex) }

{ Let's read some value... }

FUNCTION TSharedMemory.ReadValue(Offset: Integer): Byte;
VAR
  dwResultWaitMutex: DWord;
BEGIN
  IF FCreated THEN
  BEGIN
    dwResultWaitMutex := WaitForSingleObject(FHndMutex, INFINITE);

    IF dwResultWaitMutex = WAIT_OBJECT_0 THEN
    BEGIN
      Result := FP^[Offset];
      ReleaseMutex(FHndMutex);
    END
    ELSE
    BEGIN
      SetLastError(ERROR_OPEN_FAILED);
      Result := 0;
    END;
  END
  ELSE
  BEGIN
    RAISE
      ESharedMemoryError.Create('Can''t read value on a not created shared memory');
    SetLastError(ERROR_ACCESS_DENIED);
    Result := 0;
  END;
END;

{ Let's save all shared memory to file... }

PROCEDURE TSharedMemory.Flush;
VAR
  dwResultWaitMutex: DWord;
BEGIN
  IF FCreated THEN
  BEGIN
    dwResultWaitMutex := WaitForSingleObject(FHndMutex, INFINITE);

    IF dwResultWaitMutex = WAIT_OBJECT_0 THEN
    BEGIN
      IF FHndFile <> INVALID_HANDLE_VALUE THEN
        IF NOT FlushViewOfFile(FStartingDirFileView, FMemSize) THEN
          RAISE
            ESharedMemoryError.Create('Unexpected error flushing memory to file');

      ReleaseMutex(FHndMutex);
    END;
  END
  ELSE
  BEGIN
    RAISE
      ESharedMemoryError.Create('Can''t flush memory on a not created shared memory');
    SetLastError(ERROR_ACCESS_DENIED);
  END;
END;

{ Let's write an entire block... }

PROCEDURE TSharedMemory.WriteBlock(Offset: Integer; Count: DWord; Value: Byte);
VAR
  dwResultWaitMutex: DWord;
BEGIN
  IF FCreated THEN
  BEGIN
    dwResultWaitMutex := WaitForSingleObject(FHndMutex, INFINITE);

    IF dwResultWaitMutex = WAIT_OBJECT_0 THEN
    BEGIN
      FillMemory(@FP^[Offset], Count, Value);
      ReleaseMutex(FHndMutex);
    END;
  END
  ELSE
  BEGIN
    RAISE
      ESharedMemoryError.Create('Can''t write block on a not created shared memory');
    SetLastError(ERROR_ACCESS_DENIED);
  END;
END;

{ Let's write an entire block and save it to file... }

PROCEDURE TSharedMemory.WriteBlockFlush(Offset: Integer; Count: DWord; Value: Byte);
VAR
  dwResultWaitMutex: DWord;
BEGIN
  IF FCreated THEN
  BEGIN
    dwResultWaitMutex := WaitForSingleObject(FHndMutex, INFINITE);

    IF dwResultWaitMutex = WAIT_OBJECT_0 THEN
    BEGIN
      FillMemory(@FP^[Offset], Count, Value);

      IF FHndFile <> INVALID_HANDLE_VALUE THEN
        IF NOT FlushViewOfFile(FStartingDirFileView, DWord(Offset) + Count) THEN
          RAISE
            ESharedMemoryError.Create('Unexpected error flushing memory to file');

      ReleaseMutex(FHndMutex);
    END;
  END
  ELSE
  BEGIN
    RAISE
      ESharedMemoryError.Create('Can''t write block on a not created shared memory');
    SetLastError(ERROR_ACCESS_DENIED);
  END;
END;


{ This is the end for SharedMemory.pas }

end.
